package com.io7m.documents.displacement2d;

import java.io.IOException;

import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;

public final class Ripple implements Runnable
{
  private static int SCREEN_WIDTH  = 640;
  private static int SCREEN_HEIGHT = 480;

  public static void main(
    final String args[])
  {
    if (args.length < 2) {
      System.err.println("usage: image displace");
      System.exit(1);
    }

    try {
      Display.setTitle("Ripple");
      Display.setDisplayMode(new DisplayMode(
        Ripple.SCREEN_WIDTH,
        Ripple.SCREEN_HEIGHT));
      Display.create();
      final Ripple r = new Ripple(args[0], args[1]);
      r.run();
    } catch (final LWJGLException e) {
      System.err.print("fatal: LGJWL error: " + e.getMessage());
    } catch (final IOException e) {
      System.err.print("fatal: i/o error: " + e.getMessage());
    } finally {
      Display.destroy();
    }
  }

  private final int        framebuffer;
  private final int        framebuffer_texture;
  private final int        shader_uv;
  private final int        shader_displace;
  private final int        texture_image;
  private final int        texture_displacement_map;

  private final static int TEXTURE_WIDTH  = 256;
  private final static int TEXTURE_HEIGHT = 256;

  private float            time           = 0;

  Ripple(
    final String image,
    final String displace)
    throws IOException
  {
    this.texture_image = Utilities.loadTexture(image);
    this.texture_displacement_map = Utilities.loadTexture(displace);

    this.framebuffer_texture =
      Utilities.createEmptyTexture(
        Ripple.TEXTURE_WIDTH,
        Ripple.TEXTURE_HEIGHT);
    this.framebuffer = Utilities.createFramebuffer(this.framebuffer_texture);

    this.shader_uv =
      Utilities.createShader("dist/shader_uv.v", "dist/shader_uv.f");
    this.shader_displace =
      Utilities.createShader("dist/shader_uv.v", "dist/shader_displace.f");
  }

  private void render()
  {
    this.time += 1.0f;

    this.renderToTexture();
    this.renderScene();
  }

  private void renderScene()
  {
    GL11.glMatrixMode(GL11.GL_PROJECTION);
    GL11.glLoadIdentity();
    GL11.glFrustum(-1, 1, -1, 1, 1, 100);

    GL11.glMatrixMode(GL11.GL_MODELVIEW);
    GL11.glLoadIdentity();

    GL11.glTranslated(0, 0, -1.25);
    GL11.glRotated(30, 0, 0, 1);

    GL11.glViewport(0, 0, Ripple.SCREEN_WIDTH, Ripple.SCREEN_HEIGHT);
    GL11.glClearColor(0.25f, 0.25f, 0.25f, 1.0f);
    GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);

    GL20.glUseProgram(this.shader_uv);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.framebuffer_texture);
      Utilities.checkGL();

      final int ut = GL20.glGetUniformLocation(this.shader_uv, "texture");
      GL20.glUniform1i(ut, 0);
      Utilities.checkGL();

      GL11.glBegin(GL11.GL_QUADS);
      {
        GL11.glTexCoord2f(0, 0);
        GL11.glVertex3d(-0.75, 0.75, 0);
        GL11.glTexCoord2f(0, 1);
        GL11.glVertex3d(-0.75, -0.75, 0);
        GL11.glTexCoord2f(1, 1);
        GL11.glVertex3d(0.75, -0.75, 0);
        GL11.glTexCoord2f(1, 0);
        GL11.glVertex3d(0.75, 0.75, 0);
      }
      GL11.glEnd();
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  private void renderToTexture()
  {
    GL11.glMatrixMode(GL11.GL_PROJECTION);
    GL11.glLoadIdentity();
    GL11.glOrtho(0, 1, 0, 1, 1, 100);

    GL11.glMatrixMode(GL11.GL_MODELVIEW);
    GL11.glLoadIdentity();
    GL11.glTranslated(0, 0, -1);

    GL11.glViewport(0, 0, Ripple.TEXTURE_WIDTH, Ripple.TEXTURE_HEIGHT);
    GL11.glClearColor(0.25f, 0.25f, 0.25f, 1.0f);
    GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.framebuffer);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.texture_image);
      GL13.glActiveTexture(GL13.GL_TEXTURE1);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.texture_displacement_map);

      GL20.glUseProgram(this.shader_displace);
      {
        final int ut =
          GL20.glGetUniformLocation(this.shader_displace, "texture");
        final int udm =
          GL20.glGetUniformLocation(this.shader_displace, "displace_map");
        final int umax =
          GL20.glGetUniformLocation(this.shader_displace, "maximum");
        final int utime =
          GL20.glGetUniformLocation(this.shader_displace, "time");

        GL20.glUniform1i(ut, 0);
        GL20.glUniform1i(udm, 1);
        GL20.glUniform1f(umax, 0.2f);
        GL20.glUniform1f(utime, this.time);
        Utilities.checkGL();

        GL11.glBegin(GL11.GL_QUADS);
        {
          GL11.glTexCoord2f(0, 1);
          GL11.glVertex3d(0, 1, 0);
          GL11.glTexCoord2f(0, 0);
          GL11.glVertex3d(0, 0, 0);
          GL11.glTexCoord2f(1, 0);
          GL11.glVertex3d(1, 0, 0);
          GL11.glTexCoord2f(1, 1);
          GL11.glVertex3d(1, 1, 0);
        }
        GL11.glEnd();
      }
      GL20.glUseProgram(0);

      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
      GL13.glActiveTexture(GL13.GL_TEXTURE1);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
    }
    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    Utilities.checkGL();
  }

  @Override public void run()
  {
    try {
      while (Display.isCloseRequested() == false) {
        this.render();
        Display.update();
        Display.sync(60);
      }
    } finally {
      Display.destroy();
    }
  }
}
